from __future__ import annotations

import builtins
import dataclasses
import datetime
import decimal
import enum
import inspect
import ipaddress
import pathlib
import types
import uuid
from collections import namedtuple
from operator import attrgetter
from typing import (
    Any,
    Optional,
    Union,
    Collection,
    Mapping,
    ClassVar,
    Type,
    Tuple,
    TypeVar,
    Iterable,
    Iterator,
    Hashable,
    NamedTuple,
)

import typic
import typic.common
import typic.util as util
import typic.strict as strict

from typic.compat import (
    Final,
    ForwardRef,
    Literal,
    lru_cache,
    TypeGuard,
    TypedDict,
    Protocol,
)

ObjectT = TypeVar("ObjectT")
"""A type-alias for a python object.

Used in place of :py:class:`Any` for better type-hinting.

Examples
--------
>>> import typic
>>> from typing import Type
>>> def get_type(obj: typic.ObjectT) -> Type[ObjectT]:
...     return type(obj)
...
>>> get_type("")  # IDE/mypy tracks input type
str
"""

__all__ = (
    "BUILTIN_TYPES",
    "ObjectT",
    "isbuiltininstance",
    "isbuiltintype",
    "isbuiltinsubtype",
    "isclassvartype",
    "iscollectiontype",
    "isconstrained",
    "isdatetype",
    "isenumtype",
    "isfinal",
    "isforwardref",
    "isfromdictclass",
    "isfrozendataclass",
    "ishashable",
    "isinstance",
    "isiterabletype",
    "isliteral",
    "ismappingtype",
    "isnamedtuple",
    "isoptionaltype",
    "isproperty",
    "isreadonly",
    "issimpleattribute",
    "isstrict",
    "isstdlibinstance",
    "isstdlibtype",
    "isstdlibsubtype",
    "issubclass",
    "istimetype",
    "istimedeltatype",
    "istupletype",
    "istypeddict",
    "istypedtuple",
    "isuniontype",
    "isuuidtype",
    "iswriteonly",
    "should_unwrap",
)


# Here we are with a manually-defined set of builtin-types.
# This probably won't break anytime soon, but we shall see...
BuiltInTypeT = Union[
    int, bool, float, str, bytes, bytearray, list, set, frozenset, tuple, dict, None
]
BUILTIN_TYPES = frozenset(
    (type(None), *(t for t in BuiltInTypeT.__args__ if t is not None))  # type: ignore
)
BUILTIN_TYPES_TUPLE = tuple(BUILTIN_TYPES)
STDLibTypeT = Union[
    BuiltInTypeT,
    datetime.datetime,
    datetime.date,
    datetime.timedelta,
    datetime.time,
    decimal.Decimal,
    ipaddress.IPv4Address,
    ipaddress.IPv6Address,
    pathlib.Path,
    uuid.UUID,
    uuid.SafeUUID,
]
STDLIB_TYPES = frozenset(
    (type(None), *(t for t in STDLibTypeT.__args__ if t is not None))  # type: ignore
)
STDLIB_TYPES_TUPLE = tuple(STDLIB_TYPES)


@lru_cache(maxsize=None)
def isbuiltintype(obj: Type[ObjectT]) -> TypeGuard[Type[BuiltInTypeT]]:
    """Check whether the provided object is a builtin-type.

    Python stdlib and Python documentation have no "definitive list" of
    builtin-**types**, despite the fact that they are well-known. The closest we have
    is https://docs.python.org/3.7/library/functions.html, which clumps the
    builtin-types with builtin-functions. Despite clumping these types with functions
    in the documentation, these types eval as False when compared to
    :py:class:`types.BuiltinFunctionType`, which is meant to be an alias for the
    builtin-functions listed in the documentation.

    All this to say, here we are with a custom check to determine whether a type is a
    builtin.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> from typing import NewType, Mapping
    >>> typic.isbuiltintype(str)
    True
    >>> typic.isbuiltintype(NewType("MyStr", str))
    True
    >>> class Foo: ...
    ...
    >>> typic.isbuiltintype(Foo)
    False
    >>> typic.isbuiltintype(Mapping)
    False
    """
    return (
        util.resolve_supertype(obj) in BUILTIN_TYPES
        or util.resolve_supertype(type(obj)) in BUILTIN_TYPES
    )


@lru_cache(maxsize=None)
def isstdlibtype(obj: Type[ObjectT]) -> TypeGuard[Type[STDLibTypeT]]:
    return (
        util.resolve_supertype(obj) in STDLIB_TYPES
        or util.resolve_supertype(type(obj)) in STDLIB_TYPES
    )


@lru_cache(maxsize=None)
def isbuiltinsubtype(t: Type[ObjectT]) -> TypeGuard[Type[BuiltInTypeT]]:
    """Check whether the provided type is a subclass of a builtin-type.

    Parameters
    ----------
    t

    Examples
    --------
    >>> import typic
    >>> from typing import NewType, Mapping
    >>> class SuperStr(str): ...
    ...
    >>> typic.isbuiltinsubtype(SuperStr)
    True
    >>> typic.isbuiltinsubtype(NewType("MyStr", SuperStr))
    True
    >>> class Foo: ...
    ...
    >>> typic.isbuiltintype(Foo)
    False
    >>> typic.isbuiltintype(Mapping)
    False
    """
    return issubclass(util.resolve_supertype(t), BUILTIN_TYPES_TUPLE)


@lru_cache(maxsize=None)
def isstdlibsubtype(t: Type[ObjectT]) -> TypeGuard[Type[STDLibTypeT]]:
    return issubclass(util.resolve_supertype(t), STDLIB_TYPES_TUPLE)


def isbuiltininstance(o: ObjectT) -> TypeGuard[BuiltInTypeT]:
    return _isinstance(o, BUILTIN_TYPES_TUPLE)


def isstdlibinstance(o: ObjectT) -> TypeGuard[STDLibTypeT]:
    return _isinstance(o, STDLIB_TYPES_TUPLE)


@lru_cache(maxsize=None)
def isoptionaltype(obj: Type[ObjectT]) -> TypeGuard[Optional]:
    """Test whether an annotation is :py:class`typing.Optional`, or can be treated as.

    :py:class:`typing.Optional` is an alias for `typing.Union[<T>, None]`, so both are
    "optional".

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> from typing import Optional, Union, Dict
    >>> typic.isoptionaltype(Optional[str])
    True
    >>> typic.isoptionaltype(Union[str, None])
    True
    >>> typic.isoptionaltype(Dict[str, None])
    False
    """
    args = getattr(obj, "__args__", ())
    return (
        len(args) > 1
        and args[-1]
        in {
            type(None),
            None,
        }  # noqa: E721 - we don't know what args[-1] is, so this is safer
        and util.get_name(util.origin(obj)) in {"Optional", "Union", "Literal"}
    )


@lru_cache(maxsize=None)
def isuniontype(obj: Type[ObjectT]) -> TypeGuard[Union]:
    return util.get_name(util.origin(obj)) in {"Union", "UnionType"}


@lru_cache(maxsize=None)
def isreadonly(obj: Type[ObjectT]) -> TypeGuard[typic.common.ReadOnly]:
    """Test whether an annotation is marked as :py:class:`typic.ReadOnly`

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> from typing import NewType
    >>> typic.isreadonly(typic.ReadOnly[str])
    True
    >>> typic.isreadonly(NewType("Foo", typic.ReadOnly[str]))
    True
    """
    return util.origin(obj) is typic.common.ReadOnly


@lru_cache(maxsize=None)
def isfinal(obj: Type[ObjectT]) -> bool:
    """Test whether an annotation is :py:class:`typing.Final`.

    Examples
    --------

    >>> import typic
    >>> from typing import NewType
    >>> from typic.compat import Final
    >>> typic.isfinal(Final[str])
    True
    >>> typic.isfinal(NewType("Foo", Final[str]))
    True
    """
    return util.origin(obj) is Final


@lru_cache(maxsize=None)
def isliteral(obj: Type) -> TypeGuard[Literal]:
    return util.origin(obj) is Literal or (
        obj.__class__ is ForwardRef and obj.__forward_arg__.startswith("Literal")
    )


@lru_cache(maxsize=None)
def iswriteonly(obj: Type[ObjectT]) -> TypeGuard[typic.common.WriteOnly]:
    """Test whether an annotation is marked as :py:class:`typic.WriteOnly`.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> from typing import NewType
    >>> typic.iswriteonly(typic.WriteOnly[str])
    True
    >>> typic.iswriteonly(NewType("Foo", typic.WriteOnly[str]))
    True
    """
    return util.origin(obj) is typic.common.WriteOnly


@lru_cache(maxsize=None)
def isstrict(obj: Type[ObjectT]) -> TypeGuard[typic.Strict]:
    """Test whether an annotation is marked as :py:class:`typic.WriteOnly`.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> from typing import NewType
    >>> typic.isstrict(typic.Strict[str])
    True
    >>> typic.isstrict(NewType("Foo", typic.Strict[str]))
    True
    """
    return util.origin(obj) is strict.Strict


@lru_cache(maxsize=None)
def isdatetype(
    obj: Type[ObjectT],
) -> TypeGuard[Type[Union[datetime.datetime, datetime.date]]]:
    """Test whether this annotation is a a date/datetime object.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> import datetime
    >>> from typing import NewType
    >>> typic.isdatetype(datetime.datetime)
    True
    >>> typic.isdatetype(datetime.date)
    True
    >>> typic.isdatetype(NewType("Foo", datetime.datetime))
    True
    """
    return builtins.issubclass(util.origin(obj), (datetime.datetime, datetime.date))


@lru_cache(maxsize=None)
def istimetype(obj: Type[ObjectT]) -> TypeGuard[Type[datetime.time]]:
    """Test whether this annotation is a a date/datetime object.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> import datetime
    >>> from typing import NewType
    >>> typic.istimetype(datetime.time)
    True
    >>> typic.istimetype(NewType("Foo", datetime.time))
    True
    """
    return builtins.issubclass(util.origin(obj), datetime.time)


@lru_cache(maxsize=None)
def istimedeltatype(obj: Type[ObjectT]) -> TypeGuard[Type[datetime.timedelta]]:
    """Test whether this annotation is a a date/datetime object.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> import datetime
    >>> from typing import NewType
    >>> typic.istimedeltatype(datetime.timedelta)
    True
    >>> typic.istimedeltatype(NewType("Foo", datetime.timedelta))
    True
    """
    return builtins.issubclass(util.origin(obj), datetime.timedelta)


@lru_cache(maxsize=None)
def isuuidtype(obj: Type[ObjectT]) -> TypeGuard[Type[uuid.UUID]]:
    """Test whether this annotation is a a date/datetime object.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> import uuid
    >>> from typing import NewType
    >>> typic.isuuidtype(uuid.UUID)
    True
    >>> class MyUUID(uuid.UUID): ...
    ...
    >>> typic.isuuidtype(MyUUID)
    True
    >>> typic.isuuidtype(NewType("Foo", uuid.UUID))
    True
    """
    return builtins.issubclass(util.origin(obj), uuid.UUID)


_COLLECTIONS = {list, set, tuple, frozenset, dict, str, bytes}


@lru_cache(maxsize=None)
def isiterabletype(obj: Type[ObjectT]) -> TypeGuard[Type[Iterable]]:
    obj = util.origin(obj)
    return builtins.issubclass(obj, Iterable)


@lru_cache(maxsize=None)
def isiteratortype(obj: Type[ObjectT]) -> TypeGuard[Type[Iterator]]:
    obj = util.origin(obj)
    return builtins.issubclass(obj, Iterator)


@lru_cache(maxsize=None)
def istupletype(obj: Type[ObjectT]) -> TypeGuard[Type[tuple]]:
    obj = util.origin(obj)
    return obj is tuple or issubclass(obj, tuple)


@lru_cache(maxsize=None)
def iscollectiontype(obj: Type[ObjectT]) -> TypeGuard[Type[Collection]]:
    """Test whether this annotation is a subclass of :py:class:`typing.Collection`.

    Includes builtins.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> from typing import Collection, Mapping, NewType
    >>> typic.iscollectiontype(Collection)
    True
    >>> typic.iscollectiontype(Mapping[str, str])
    True
    >>> typic.iscollectiontype(str)
    True
    >>> typic.iscollectiontype(list)
    True
    >>> typic.iscollectiontype(NewType("Foo", dict))
    True
    >>> typic.iscollectiontype(int)
    False
    """
    obj = util.origin(obj)
    return obj in _COLLECTIONS or builtins.issubclass(obj, Collection)


@lru_cache(maxsize=None)
def ismappingtype(obj: Type[ObjectT]) -> TypeGuard[Type[Mapping]]:
    """Test whether this annotation is a subtype of :py:class:`typing.Mapping`.

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> from typing import Mapping, Dict, DefaultDict, NewType
    >>> typic.ismappingtype(Mapping)
    True
    >>> typic.ismappingtype(Dict[str, str])
    True
    >>> typic.ismappingtype(DefaultDict)
    True
    >>> typic.ismappingtype(dict)
    True
    >>> class MyDict(dict): ...
    ...
    >>> typic.ismappingtype(MyDict)
    True
    >>> class MyMapping(Mapping): ...
    ...
    >>> typic.ismappingtype(MyMapping)
    True
    >>> typic.ismappingtype(NewType("Foo", dict))
    True
    """
    obj = util.origin(obj)
    return builtins.issubclass(obj, dict) or builtins.issubclass(obj, Mapping)


@lru_cache(maxsize=None)
def isenumtype(obj: Type[ObjectT]) -> TypeGuard[Type[enum.Enum]]:
    """Test whether this annotation is a subclass of :py:class:`enum.Enum`

    Parameters
    ----------
    obj

    Examples
    --------

    >>> import typic
    >>> import enum
    >>>
    >>> class FooNum(enum.Enum): ...
    ...
    >>> typic.isenumtype(FooNum)
    True
    """
    return issubclass(obj, enum.Enum)


@lru_cache(maxsize=None)
def isclassvartype(obj: Type[ObjectT]) -> TypeGuard[ClassVar]:
    """Test whether an annotation is a ClassVar annotation.

    Examples
    --------
    >>> import typic
    >>> from typing import ClassVar, NewType
    >>> typic.isclassvartype(ClassVar[str])
    True
    >>> typic.isclassvartype(NewType("Foo", ClassVar[str]))
    True
    """
    obj = util.resolve_supertype(obj)
    return getattr(obj, "__origin__", obj) is ClassVar


_UNWRAPPABLE = (
    isclassvartype,
    isoptionaltype,
    isreadonly,
    iswriteonly,
    isfinal,
    isstrict,
)


@lru_cache(maxsize=None)
def should_unwrap(obj: Type[ObjectT]) -> bool:
    """Test whether we should use the __args__ attr for resolving the type.

    This is useful for determining what type to use at run-time for coercion.
    """
    return (not isliteral(obj)) and any(x(obj) for x in _UNWRAPPABLE)


@lru_cache(maxsize=None)
def isfromdictclass(obj: Type[ObjectT]) -> TypeGuard[_FromDict]:
    """Test whether this annotation is a class with a `from_dict()` method."""
    return inspect.isclass(obj) and hasattr(obj, "from_dict")


class _FromDict(Protocol):
    def from_dict(self, *args, **kwargs) -> _FromDict:
        ...


@lru_cache(maxsize=None)
def isfrozendataclass(obj: Type[ObjectT]) -> TypeGuard[_FrozenDataclass]:
    """Test whether this is a dataclass and whether it's frozen."""
    return getattr(getattr(obj, "__dataclass_params__", None), "frozen", False)


class _FrozenDataclass(Protocol):
    __dataclas_params__: dataclasses._DataclassParams  # type: ignore


_isinstance = isinstance


@lru_cache(maxsize=None)
def _type_check(t) -> bool:
    if _isinstance(t, tuple):
        return all(_type_check(x) for x in t)
    return inspect.isclass(t)


def isinstance(o: Any, t: Union[Type, Tuple[Type, ...]]) -> bool:
    """A safer instance check...

    Validates that `t` is not an instance.

    Parameters
    ----------
    o
        The object to test.
    t
        The type(s) to test against.

    Examples
    --------
    >>> import typic
    >>> typic.isinstance("", str)
    True
    >>> typic.isinstance("", "")
    False
    """
    return _type_check(t) and builtins.isinstance(o, t)


def issubclass(o: Type[Any], t: Union[Type, Tuple[Type, ...]]) -> bool:
    """A safer subclass check...

    Validates that `t` and/or `o` are not instances.

    Parameters
    ----------
    o
        The type to validate
    t
        The type(s) to validate against

    Notes
    -----
    Not compatible with classes from :py:mod:`typing`, as they return False with
    :py:func:`inspect.isclass`

    Examples
    --------
    >>> import typic
    >>> class MyStr(str): ...
    ...
    >>> typic.issubclass(MyStr, str)
    True
    >>> typic.issubclass(MyStr(), str)
    False
    >>> typic.issubclass(MyStr, str())
    False
    """
    return _type_check(t) and _type_check(o) and builtins.issubclass(o, t)


@lru_cache(maxsize=None)
def isconstrained(obj: Type[ObjectT]) -> TypeGuard[_Constrained]:
    """Test whether a type is restricted.

    Parameters
    ----------
    obj

    Examples
    --------
    >>> import typic
    >>> from typing import NewType
    >>>
    >>> @typic.constrained(ge=0, le=1)
    ... class TinyInt(int): ...
    ...
    >>> typic.isconstrained(TinyInt)
    True
    >>> Switch = NewType("Switch", TinyInt)
    >>> typic.isconstrained(Switch)
    True
    """
    return hasattr(util.resolve_supertype(obj), "__constraints__")


class _Constrained(Protocol):
    __constraints__: typic.ConstraintsProtocolT


__hashgetter = attrgetter("__hash__")


def ishashable(obj: ObjectT) -> TypeGuard[Hashable]:
    """Check whether an object is hashable.

    An order of magnitude faster than :py:class:`isinstance` with
    :py:class:`typing.Hashable`

    Parameters
    ----------
    obj

    Examples
    --------
    >>> import typic
    >>> typic.ishashable(str())
    True
    >>> typic.ishashable(frozenset())
    True
    >>> typic.ishashable(list())
    False
    """
    return __hashgetter(obj) is not None


@lru_cache(maxsize=None)
def istypeddict(obj: Type[ObjectT]) -> TypeGuard[Type[TypedDict]]:
    """Check whether an object is a :py:class:`typing.TypedDict`.

    Parameters
    ----------
    obj

    Examples
    --------
    >>> import typic
    >>> from typic.compat import TypedDict
    >>>
    >>> class FooMap(TypedDict):
    ...     bar: str
    ...
    >>> typic.istypeddict(FooMap)
    True
    """
    return (
        inspect.isclass(obj)
        and dict in {*inspect.getmro(obj)}
        and hasattr(obj, "__total__")
    )


@lru_cache(maxsize=None)
def istypedtuple(obj: Type[ObjectT]) -> TypeGuard[Type[NamedTuple]]:
    """Check whether an object is a "typed" tuple (:py:class:`typing.NamedTuple`).

    Parameters
    ----------
    obj

    Examples
    --------
    >>> import typic
    >>> from typing import NamedTuple
    >>>
    >>> class FooTup(NamedTuple):
    ...     bar: str
    ...
    >>> typic.istypedtuple(FooTup)
    True
    """
    return (
        inspect.isclass(obj)
        and issubclass(obj, tuple)
        and hasattr(obj, "__annotations__")
    )


@lru_cache(maxsize=None)
def isnamedtuple(obj: Type[ObjectT]) -> TypeGuard[namedtuple]:
    """Check whether an object is a "named" tuple (:py:func:`collections.namedtuple`).

    Parameters
    ----------
    obj

    Examples
    --------
    >>> import typic
    >>> from collections import namedtuple
    >>>
    >>> FooTup = namedtuple("FooTup", ["bar"])
    >>> typic.isnamedtuple(FooTup)
    True
    """
    return inspect.isclass(obj) and issubclass(obj, tuple) and hasattr(obj, "_fields")


def isforwardref(obj: Any) -> TypeGuard[ForwardRef]:
    return obj.__class__ is ForwardRef


def isproperty(obj) -> TypeGuard[types.GetSetDescriptorType]:
    return obj.__class__.__name__ in {"property", "cached_property"}


_ATTR_CHECKS = (inspect.isclass, inspect.isroutine, isproperty)


def issimpleattribute(v) -> bool:
    return not any(c(v) for c in _ATTR_CHECKS)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# flake8: noqa

from .array import (
    ListConstraints,
    TupleConstraints,
    SetContraints,
    FrozenSetConstraints,
    Array,
    ArrayConstraints,
)
from .common import (
    ValidatorT,
    ValidateT,
    BaseConstraints,
    ConstraintsProtocolT,
    VT,
    MultiConstraints,
    TypeConstraints,
    LiteralConstraints,
    EnumConstraints,
)
from .error import ConstraintValueError, ConstraintSyntaxError
from .mapping import DictConstraints, MappingConstraints, ObjectConstraints
from .number import DecimalContraints, FloatContraints, IntContraints, NumberT
from .text import BytesConstraints, StrConstraints
from .factory import ConstraintsT, get_constraints

__all__ = (
    "BytesConstraints",
    "ConstraintValueError",
    "ConstraintSyntaxError",
    "ConstraintsProtocolT",
    "DecimalContraints",
    "DictConstraints",
    "EnumConstraints",
    "FloatContraints",
    "FrozenSetConstraints",
    "IntContraints",
    "ListConstraints",
    "LiteralConstraints",
    "NumberT",
    "SetContraints",
    "StrConstraints",
    "TupleConstraints",
    "ValidatorT",
    "BaseConstraints",
    "get_constraints",
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['typic',
 'typic.constraints',
 'typic.ext',
 'typic.ext.schema',
 'typic.serde',
 'typic.types']

package_data = \
{'': ['*']}

install_requires = \
['future-typing>=0.4.1,<0.5.0', 'inflection>=0.5,<0.6', 'pendulum>=2.1,<3.0']

extras_require = \
{'benchmarks': ['sqlalchemy>=1.3.13,<2.0.0',
                'pytest-benchmark[histogram]>=3.2,<4.0',
                'marshmallow>=3.2,<4.0',
                'toastedmarshmallow>=2.15,<3.0',
                'djangorestframework>=3.10,<4.0',
                'pydantic[email]>=1.0,<2.0',
                'pydantic[email]>=1.0,<2.0',
                'django>=2.2,<3.0'],
 'docs': ['mkdocs>=1.1,<2.0',
          'mkdocs-material>=5.5,<6.0',
          'mkdocs-awesome-pages-plugin>=2.2.1,<3.0.0',
          'pymdown-extensions>=7.0,<8.0'],
 'json': ['ujson>=2.0'],
 'lint': ['flake8>=3.7.9,<4.0.0',
          'mypy>=0.910,<0.911',
          'black>=21,<22',
          'types-python-dateutil',
          'types-setuptools',
          'types-toml',
          'types-typed-ast',
          'types-ujson'],
 'schema': ['fastjsonschema>=2.14,<3.0'],
 'tests': ['fastjsonschema>=2.14,<3.0',
           'ujson>=2.0',
           'pytest>=6.2,<7.0',
           'pytest-cov>=2.8,<3.0',
           'sqlalchemy>=1.3.13,<2.0.0',
           'pydantic[email]>=1.0,<2.0',
           'mypy>=0.910,<0.911'],
 'tests:python_version >= "3.7.1" and python_version <= "3.9"': ['pandas>=1.1.3,<2.0.0']}

setup_kwargs = {
    'name': 'typical',
    'version': '2.6.0',
    'description': "Typical: Python's Typing Toolkit.",
    'long_description': '# typical: Python\'s Typing Toolkit\n[![image](https://img.shields.io/pypi/v/typical.svg)](https://pypi.org/project/typical/)\n[![image](https://img.shields.io/pypi/l/typical.svg)](https://pypi.org/project/typical/)\n[![image](https://img.shields.io/pypi/pyversions/typical.svg)](https://pypi.org/project/typical/)\n[![image](https://img.shields.io/github/languages/code-size/seandstewart/typical.svg?style=flat)](https://github.com/seandstewart/typical)\n[![Test & Lint](https://github.com/seandstewart/typical/workflows/Test%20&%20Lint/badge.svg)](https://github.com/seandstewart/typical/actions)\n[![Coverage](https://codecov.io/gh/seandstewart/typical/branch/master/graph/badge.svg)](https://codecov.io/gh/seandstewart/typical)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![Netlify Status](https://api.netlify.com/api/v1/badges/982a0ced-bb7f-4391-87e8-1957071d2f66/deploy-status)](https://app.netlify.com/sites/typical-python/deploys)\n\n![How Typical](static/typical.png)\n\n## Introduction\n\nTypical is a library devoted to runtime analysis, inference,\nvalidation, and enforcement of Python types,\n[PEP 484](https://www.python.org/dev/peps/pep-0484/) Type Hints, and\ncustom user-defined data-types.\n\nTypical is fully compliant with the following Python Typing PEPs:\n\n- [PEP 484 -- Type Hints](https://www.python.org/dev/peps/pep-0484/)\n- [PEP 563 -- Postponed Evaluation of Annotations](https://www.python.org/dev/peps/pep-0563/)\n- [PEP 585 -- Type Hinting Generics In Standard Collections](https://www.python.org/dev/peps/pep-0585/)\n- [PEP 586 -- Literal Types](https://www.python.org/dev/peps/pep-0586/)\n- [PEP 589 -- TypedDict: Type Hints for Dictionaries with a Fixed Set of Keys](https://www.python.org/dev/peps/pep-0589/)\n- [PEP 604 -- Allow writing union types as X | Y](https://www.python.org/dev/peps/pep-0604/)\n\nIt provides a high-level Protocol API, Functional API, and Object API to suit most any\noccasion.\n\n## Getting Started\n\nInstallation is as simple as `pip install -U typical`.\n## Help\n\nThe latest documentation is hosted at\n[python-typical.org](https://python-typical.org/).\n\n> Starting with version 2.0, All documentation is hand-crafted\n> markdown & versioned documentation can be found at typical\'s\n> [Git Repo](https://github.com/seandstewart/typical/tree/master/docs).\n> (Versioned documentation is still in-the-works directly on our\n> domain.)\n\n## A Typical Use-Case\n\nThe decorator that started it all:\n\n### `typic.al(...)`\n\n```python\nimport typic\n\n\n@typic.al\ndef hard_math(a: int, b: int, *c: int) -> int:\n    return a + b + sum(c)\n\nhard_math(1, "3")\n#> 4\n\n\n@typic.al(strict=True)\ndef strict_math(a: int, b: int, *c: int) -> int:\n    return a + b + sum(c)\n\nstrict_math(1, 2, 3, "4")\n#> Traceback (most recent call last):\n#>  ...\n#> typic.constraints.error.ConstraintValueError: Given value <\'4\'> fails constraints: (type=int, nullable=False, coerce=False)\n  \n```\n\nTypical has both a high-level *Object API* and high-level\n*Functional API*. In general, any method registered to one API is also\navailable to the other.\n\n### The Protocol API\n\n```python\nimport dataclasses\nfrom typing import Iterable\n\nimport typic\n\n\n@typic.constrained(ge=1)\nclass ID(int):\n    ...\n\n\n@typic.constrained(max_length=280)\nclass Tweet(str):\n    ...\n\n\n@dataclasses.dataclass # or typing.TypedDict or typing.NamedTuple or annotated class...\nclass Tweeter:\n    id: ID\n    tweets: Iterable[Tweet]\n\n\njson = \'{"id":1,"tweets":["I don\\\'t understand Twitter"]}\'\nprotocol = typic.protocol(Tweeter)\n\nt = protocol.transmute(json)\nprint(t)\n#> Tweeter(id=1, tweets=["I don\'t understand Twitter"])\n\nprint(protocol.tojson(t))\n#> \'{"id":1,"tweets":["I don\\\'t understand Twitter"]}\'\n\nprotocol.validate({"id": 0, "tweets": []})\n#> Traceback (most recent call last):\n#>  ...\n#> typic.constraints.error.ConstraintValueError: Tweeter.id: value <0> fails constraints: (type=int, nullable=False, coerce=False, ge=1)\n```\n\n### The Functional API\n\n```python\nimport dataclasses\nfrom typing import Iterable\n\nimport typic\n\n\n@typic.constrained(ge=1)\nclass ID(int):\n    ...\n\n\n@typic.constrained(max_length=280)\nclass Tweet(str):\n    ...\n\n\n@dataclasses.dataclass # or typing.TypedDict or typing.NamedTuple or annotated class...\nclass Tweeter:\n    id: ID\n    tweets: Iterable[Tweet]\n\n\njson = \'{"id":1,"tweets":["I don\\\'t understand Twitter"]}\'\n\nt = typic.transmute(Tweeter, json)\nprint(t)\n#> Tweeter(id=1, tweets=["I don\'t understand Twitter"])\n\nprint(typic.tojson(t))\n#> \'{"id":1,"tweets":["I don\\\'t understand Twitter"]}\'\n\ntypic.validate(Tweeter, {"id": 0, "tweets": []})\n#> Traceback (most recent call last):\n#>  ...\n#> typic.constraints.error.ConstraintValueError: Tweeter.id: value <0> fails constraints: (type=int, nullable=False, coerce=False, ge=1)\n```\n\n### The Object API\n\n```python\nfrom typing import Iterable\n\nimport typic\n\n\n@typic.constrained(ge=1)\nclass ID(int):\n    ...\n\n\n@typic.constrained(max_length=280)\nclass Tweet(str):\n    ...\n\n\n@typic.klass\nclass Tweeter:\n    id: ID\n    tweets: Iterable[Tweet]\n    \n\njson = \'{"id":1,"tweets":["I don\\\'t understand Twitter"]}\'\nt = Tweeter.transmute(json)\n\nprint(t)\n#> Tweeter(id=1, tweets=["I don\'t understand Twitter"])\n\nprint(t.tojson())\n#> \'{"id":1,"tweets":["I don\\\'t understand Twitter"]}\'\n\nTweeter.validate({"id": 0, "tweets": []})\n#> Traceback (most recent call last):\n#>  ...\n#> typic.constraints.error.ConstraintValueError: Given value <0> fails constraints: (type=int, nullable=False, coerce=False, ge=1)\n```\n\n\n## Changelog\n\nSee our\n[Releases](https://github.com/seandstewart/typical/releases).\n',
    'author': 'Sean Stewart',
    'author_email': 'sean_stewart@me.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/seandstewart/typical',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import dataclasses
import re
from collections import defaultdict
from types import MappingProxyType
from typing import Dict, List, ClassVar, Pattern, Match, Mapping, Set, Optional
from urllib import parse

from typic.util import cached_property, slotted
from .secret import SecretStr

__all__ = (
    "NetAddrInfo",
    "NetworkAddress",
    "NetworkAddressValueError",
    "AbsoluteURL",
    "AbsoluteURLValueError",
    "HostName",
    "HostNameValueError",
    "RelativeURL",
    "RelativeURLValueError",
    "URL",
)


# By no means an exhaustive list, but a decent chunk of use-cases
DEFAULT_PORTS = defaultdict(
    set,
    {
        "http": {80},
        "https": {443},
        "ws": {80},
        "wss": {443},
        "smtp": {25},
        "ftp": {20, 21},
        "telnet": {23},
        "imap": {143},
        "rdp": {3389},
        "ssh": {25},
        "dns": {53},
        "dhcp": {67, 68},
        "pop3": {110},
        "mysql": {3306},
        "vertica": {5434},
        "postgresql": {5432},
    },
)
NET_ADDR_PATTERN = re.compile(
    r"""
    ^
    (
        # Scheme
        ((?P<scheme>(?:[a-z0-9\.\-\+]*))://)?
        # Auth
        (?P<auth>(?:(?P<username>[^:@]+?)[:@](?P<password>[^:@]*?)[:@]))?
        # Host
        (?P<host>(?:
            # Domain
            (?P<domain>
                (?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+
                (?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)
            )
            # Localhost
            |(?P<localhost>localhost)
            |(?P<dotless>(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.?))
            # IPV4
            |(?P<ipv4>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})
            # IPV6
            |(?P<ipv6>\[[A-F0-9]*:[A-F0-9:]+\])
        ))?
        # Port
        (:(?P<port>(?:\d+)))?
    )?
    # Path, Q-string & fragment
    (?P<relative>(?:/?|[/?#]\S+))
    $
    """,
    re.IGNORECASE | re.VERBOSE,
)
PRIVATE_HOSTS = {"localhost", "127.0.0.1"}
INTERNAL_HOSTS = PRIVATE_HOSTS | {"0.0.0.0"}


INTERNAL_IP_PATTERN = re.compile(
    r"""
    ^
    # IPv4
    (127\.)|
    (192\.168\.)|
    (10\.)|(172\.1[6-9]\.)|
    (172\.2[0-9]\.)|(172\.3[0-1]\.)|
    # IPv6
    (::1)|([F][CD])
    $
    """,
    re.I | re.VERBOSE,
)


class NetworkAddressValueError(ValueError):
    """A generic error indicating the value is not a valid network address."""


@slotted(dict=True)
@dataclasses.dataclass(frozen=True)
class NetAddrInfo:
    """Detailed information about a network address.

    Can be called directly, generated by casting a :py:class:`str` as
    :py:class:`NetworkAddress`, or created with :py:meth:`NetAddrInfo.from_str`
    """

    scheme: str
    """The net-address scheme, e.g., `http`, `tcp`, `ssh`, etc."""
    auth: str
    """The user auth info."""
    password: SecretStr
    """The user's password."""
    host: str
    """The host for this addres, e.g. `0.0.0.0`, `foobar.net`."""
    port: int
    """The port for this net-address"""
    path: str
    """The URI path."""
    qs: str
    """The query-string, unparsed, e.g. `?id=1&name=foo`"""
    params: str
    """The url parameters, unparsed, e.g. `id=2;foo=bar`"""
    fragment: str
    """The uri fragment, e.g. `#some-page-anchor`"""
    is_ip: bool = False

    PATTERN: ClassVar[Pattern] = NET_ADDR_PATTERN
    DEFAULT_PORTS: ClassVar[Dict] = DEFAULT_PORTS
    PRIVATE_HOSTS: ClassVar[Set[str]] = PRIVATE_HOSTS
    INTERNAL_HOSTS: ClassVar[Set[str]] = INTERNAL_HOSTS

    @classmethod
    def from_str(cls, value) -> "NetAddrInfo":
        """Parse a string, validate, and return an instance of :py:class:`NetAddrInfo`."""
        match: Optional[Match] = cls.PATTERN.match(value)
        if not match or not value:
            raise NetworkAddressValueError(f"{value!r} is not a valid network address.")
        scheme, host = match["scheme"] or "", match["host"] or ""
        if scheme and not host:
            raise NetworkAddressValueError(f"{value!r} is not a valid network address.")
        # why re-invent the wheel here? this is fast and correct.
        parsed: parse.ParseResult = parse.urlparse(match["relative"] or "")
        # get/set the port
        port = int(match["port"] or 0)
        if port == 0 and cls.DEFAULT_PORTS[scheme]:
            port = cls.DEFAULT_PORTS[scheme].copy().pop()

        return cls(
            scheme=scheme,
            auth=match["auth"] or "",
            password=SecretStr(match["password"] or ""),
            host=host,
            port=port,
            path=parsed.path,
            qs=parsed.query,
            params=parsed.params,
            fragment=parsed.fragment,
            is_ip=bool(match["ipv4"] or match["ipv6"]),
        )

    @cached_property
    def base(self) -> str:
        """The 'base' of the URL, including scheme, auth, and host."""
        base = f"{self.scheme}://" if self.scheme else ""
        port = f":{self.port}" if self.port and not self.is_default_port else ""
        return f"{base}{self.auth}{self.host}{port}"

    @cached_property
    def relative(self):
        """The 'relative' portion of the URL: path, params, query, and fragment."""
        params = f";{self.params}" if self.params else ""
        qs = f"?{self.qs}" if self.qs else ""
        fragment = f"#{self.fragment}" if self.fragment else ""
        return f"{self.path}{params}{qs}{fragment}"

    @cached_property
    def address(self) -> str:
        """The fully-qualified network address.

        If this instance was generated from a string, it will match."""
        return f"{self.base}{self.relative}"

    @cached_property
    def address_encoded(self) -> str:
        """The fully-qualified network address, encoded."""
        return parse.quote(self.address)  # type: ignore

    @cached_property
    def query(self) -> Mapping[str, List[str]]:
        """The query-string, parsed into a mapping of key -> [values, ...]."""
        return MappingProxyType(parse.parse_qs(self.qs) if self.qs else {})

    @cached_property
    def parameters(self) -> Mapping[str, List[str]]:
        """The params, parsed into a mapping of key -> [values, ...]."""
        return MappingProxyType(parse.parse_qs(self.params) if self.params else {})

    @cached_property
    def is_default_port(self) -> bool:
        """Whether address is using the default port assigned to the given scheme."""
        defaults = DEFAULT_PORTS[self.scheme] | {0}
        return self.port in defaults

    @cached_property
    def is_relative(self) -> bool:
        """Whether address is 'relative' (i.e., whether a scheme is provided)."""
        return not self.scheme

    @cached_property
    def is_absolute(self) -> bool:
        """The opposite of `is_relative`."""
        return not self.is_relative

    @cached_property
    def is_private(self) -> bool:
        """Whether or not the URL is using a 'private' host, i.e., 'localhost'."""
        return self.host in PRIVATE_HOSTS

    @cached_property
    def is_internal(self) -> bool:
        """Whether the host provided is an 'internal' host.

        This may or may not be private, hence the distinction."""
        return bool(
            self.host
            and self.host in INTERNAL_HOSTS
            or (self.is_ip and INTERNAL_IP_PATTERN.match(self.host))
        )


# Deepcopy is broken for frozen dataclasses with slots.
# https://github.com/python/cpython/pull/17254
# NetAddrInfo.__slots__ = tuple(
#     _.name for _ in dataclasses.fields(NetAddrInfo)
# )


class NetworkAddress(str):
    """A generic, immutable network address string.

    Detailed information about the network address string can be looked up via
    :py:attr:`NetworkAddress.info`.

    This object is the base object for network-related objects.
    :py:class:`URL` has a much richer interface.

    Examples
    --------

    >>> import typic
    >>> net_addr = typic.NetworkAddress("http://foo.bar/bazz;foo=bar?buzz=1#loc")
    >>> net_addr.info.is_absolute
    True
    >>> net_addr.info.host
    'foo.bar'
    >>> net_addr.info.scheme
    'http'
    >>> net_addr.info.address_encoded
    'http%3A//foo.bar/bazz%3Bfoo%3Dbar%3Fbuzz%3D1%23loc'
    >>> net_addr.info.query
    mappingproxy({'buzz': ['1']})
    >>> net_addr.info.parameters
    mappingproxy({'foo': ['bar']})
    >>> net_addr.info.fragment
    'loc'
    >>> domain = typic.URL("foo.bar")
    >>> domain.info.is_relative
    True
    >>> domain.info.host
    'foo.bar'
    >>> net_addr
    'http://foo.bar/bazz;foo=bar?buzz=1#loc'
    >>> import json
    >>> json.dumps([net_addr])
    '["http://foo.bar/bazz;foo=bar?buzz=1#loc"]'

    See Also
    --------
    :py:class:`NetAddrInfo`
    :py:class:`URL`

    Notes
    -----
    This object inherits directly from :py:class:`str` and so is natively
    JSON-serializable.
    """

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        # Initialize the info so we get validation immediately.
        v.info
        return v

    def __setattr__(self, key, value):
        raise AttributeError(
            f"attempting to set attribute on immutable type {type(self)}"
        )

    def __delattr__(self, key):
        raise AttributeError(
            f"attempting to delete attribute on immutable type {type(self)}"
        )

    @cached_property
    def info(self) -> NetAddrInfo:
        return NetAddrInfo.from_str(self)


class URLValueError(NetworkAddressValueError):
    """Generic error for an invalid value passed to URL."""

    pass


class URL(NetworkAddress):
    """A string which parses the value provided as if it were a URL.

    Detailed information about the url string can be looked up via :py:attr:`URL.info`.

    Examples
    --------

    >>> import typic
    >>> url = typic.URL("http://foo.bar/bazz")
    >>> more = url / 'foo' / 'bar'
    >>> more
    'http://foo.bar/bazz/foo/bar'
    >>> typic.URL(url.info.base) / 'other'
    'http://foo.bar/other'

    See Also
    --------
    :py:class:`NetworkAddress`
    :py:class:`NetAddrInfo`

    Notes
    -----
    This object inherits directly from :py:class:`NetworkAddress` and so is natively
    JSON-serializable.
    """

    def join(self, other) -> "URL":
        """Join another URL with this one.

        This works roughly like :py:meth:`pathlib.Path.joinpath`.

        Unlike :py:func:`urllib.parse.urljoin`, this method allows the user to build
        onto existing paths.
        """
        cls = type(self)
        # best guess at the path
        other_info: NetAddrInfo = cls(other).info  # type: ignore
        self_info: NetAddrInfo = self.info  # type: ignore
        other = (other_info.path or parse.urlparse(other).path or "").lstrip("/")
        other = f"{self_info.path.rstrip('/') or ''}/{other}"
        return cls(parse.urljoin(self_info.base, other))  # type: ignore

    def __truediv__(self, other) -> "URL":
        """Overloading some operators to make it easier. Uses `:py:meth:`URL.join`."""
        return self.join(other)

    def __rtruediv__(self, other) -> "URL":
        return URL(other) / self


class AbsoluteURLValueError(URLValueError):
    pass


class AbsoluteURL(URL):
    """An absolute URL.

    See Also
    --------
    :py:class:`URL`
    """

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        if v.info.is_relative:
            raise AbsoluteURLValueError(f"<{v!r}> is not an absolute URL.") from None
        return v


class RelativeURLValueError(URLValueError):
    pass


class RelativeURL(URL):
    """A relative URL.

    See Also
    --------
    :py:class:`URL`
    """

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        if v.info.is_absolute:
            raise RelativeURLValueError(f"<{v!r}> is not a relative URL.") from None
        return v


class HostNameValueError(NetworkAddressValueError):
    pass


class HostName(NetworkAddress):
    """A network address referencing only a host-name (e.g. foo.bar.com).

    See Also
    --------
    :py:class:`NetworkAddress`
    :py:class:`NetAddrInfo`

    Notes
    -----
    This object inherits directly from :py:class:`NetworkAddress` and, so is natively
    JSON-serializable.
    """

    def __new__(cls, *args, **kwargs):
        v = super().__new__(cls, *args, **kwargs)
        if not v.info.host or any((v.info.scheme, v.info.auth, v.info.relative)):
            raise HostNameValueError(f"<{v!r}> is not a hostname.") from None
        return v



from __future__ import annotations

import dataclasses
import inspect
import reprlib
import sys
import warnings
from types import FrameType
from typing import (
    Callable,
    Union,
    Type,
    Any,
    Tuple,
    Mapping,
    Deque,
    Optional,
    Dict,
    Iterable,
    cast,
    TypeVar,
    AnyStr,
    Iterator,
    TYPE_CHECKING,
    Generic,
)

from typic import strict as st, util, constraints as const
from typic.checks import isclassvartype
from typic.common import AnyOrTypeT, Case, EMPTY, ObjectT, OriginT
from typic.compat import TypedDict, ForwardRef, evaluate_forwardref, Protocol
from typic.types import freeze

if TYPE_CHECKING:  # pragma: nocover
    from .resolver import Resolver


@util.slotted(dict=False)
@dataclasses.dataclass(unsafe_hash=True)
class SerdeProtocol(Generic[OriginT]):
    """An actionable run-time serialization & deserialization protocol for a type."""

    annotation: Annotation[Type[OriginT]]
    """The target annotation and various meta-data."""
    constraints: Optional[const.ConstraintsProtocolT[OriginT]]
    """Type restriction configuration, if any."""
    deserialize: DeserializerT[OriginT] = dataclasses.field(repr=False)
    """The callable to deserialize data into the annotation."""
    decode: DecoderT[OriginT] = dataclasses.field(repr=False)
    """Decode an input from the on-the-wire format to the annotation."""
    serialize: SerializerT[OriginT] = dataclasses.field(repr=False)
    """The callable to serialize an instance of the annotation."""
    encode: EncoderT[OriginT] = dataclasses.field(repr=False)
    """Encode an instance of the annotation into the provided on-the-wire format."""
    validate: const.ValidateT[OriginT] = dataclasses.field(repr=False)
    """Validate an input against the annotation."""
    translate: TranslatorT[OriginT] = dataclasses.field(repr=False)
    """Translate an instance of the annotation into another type."""
    iterate: FieldIteratorT = dataclasses.field(repr=False)
    """Iterate over an instance of the annotation, if possible."""
    tojson: EncoderT[OriginT] = dataclasses.field(repr=False)
    """Dump an instance of the annotation to valid JSON."""
    transmute: DeserializerT[OriginT] = dataclasses.field(repr=False, init=False)
    """Transmute an input into the annotation."""
    primitive: SerializerT[OriginT] = dataclasses.field(repr=False, init=False)
    """Get the "primitive" representation of the annotation."""

    def __post_init__(self):
        # Pin the transmuter and the primitiver
        self.transmute = self.deserialize
        self.primitive = self.serialize

    def __call__(self, val: ObjectT) -> OriginT:
        return self.transmute(val)  # type: ignore


_OutputT = TypeVar("_OutputT", covariant=True)
_InputT = TypeVar("_InputT", contravariant=True)


class EncoderT(Protocol[_InputT]):
    """The signature of an on-the-wire encoder for an output."""

    __name__: str
    __qualname__: str

    def __call__(self, value: _InputT, **kwargs) -> AnyStr:
        ...


class DecoderT(Protocol[_OutputT]):
    """The signature of an on-the-wire decoder for an input."""

    __name__: str
    __qualname__: str

    def __call__(self, value: AnyStr, **kwargs) -> _OutputT:
        ...


class TranslatorT(Protocol[_InputT]):
    """The signature of a type translator pinned to the origin type of `InputT`."""

    __name__: str
    __qualname__: str

    def __call__(self, value: _InputT, target: Type[_OutputT]) -> _OutputT:
        ...


class SerializerT(Protocol[_InputT]):
    """The signature of a type serializer."""

    __name__: str
    __qualname__: str

    def __call__(
        self, obj: _InputT, *, lazy: bool = False, name: util.ReprT = None
    ) -> Union[PrimitiveT, Iterator[PrimitiveT]]:
        ...


class DeserializerT(Protocol[_OutputT]):
    """The signature of a type deserializer."""

    __name__: str
    __qualname__: str

    def __call__(self, val: Any) -> _OutputT:
        ...


class FieldIteratorT(Protocol[_InputT]):
    """The type-signature for a FieldIterator function."""

    __name__: str
    __qualname__: str

    def __call__(
        self, o: _InputT, *, values: bool = False, **kwargs
    ) -> Iterator[Union[Tuple[str, Any], Any]]:
        ...


PrimitiveT = TypeVar("PrimitiveT", str, int, float, list, dict)
""""""
OmitSettingsT = Tuple[AnyOrTypeT, ...]
"""Specify types or values which you wish to omit from the output."""
FieldSerializersT = Mapping[str, SerializerT]
"""A mapping of field names to their serializer functions."""
FieldDeserializersT = Mapping[str, DeserializerT]
"""A mapping of field names to their deserializer functions."""

DeserializerTypeCheckT = Callable[[Type[Any]], bool]
"""A type alias for the expected signature of a type-check for a coercer.

Type-checkers should return a boolean indicating whether the provided type is valid for
a given coercer.
"""
DeserializerRegistryT = Deque[Tuple[DeserializerTypeCheckT, DeserializerT]]
"""A stack of potential deserializers and their function identifiers."""

FieldSettingsT = Union[Tuple[str, ...], Mapping[str, str]]
"""An iterable of fields to ensure are included on *out* and retrieved on *in*

A mapping should be of attribute name -> out/in field name.
"""


@util.slotted(dict=False)
@dataclasses.dataclass(unsafe_hash=True)
class SerdeFlags:
    """Optional settings for a Ser-ialization/de-serialization protocol."""

    signature_only: bool = False
    """Restrict the output of serialization to the class signature."""
    case: Optional[Case] = None
    """Select the case-style for the input/output fields."""
    omit: Optional[OmitSettingsT] = None
    """Provide a tuple of types or values which should be omitted on serialization."""
    fields: Optional[FieldSettingsT] = None
    """Ensure a set of fields are included in the output.

    If given a mapping, provide a mapping to the output field name.
    """
    exclude: Optional[Iterable[str]] = None
    """Provide a set of fields which will be excluded from the output."""
    encoder: Optional[EncoderT] = None
    """Provide a callable which can encode your data to a bytes/binary output."""
    decoder: Optional[DecoderT] = None
    """Provide a callable with can decode a bytes/binary input for deserialization."""

    def __init__(
        self,
        *,
        signature_only: bool = False,
        case: Case = None,
        omit: OmitSettingsT = None,
        fields: FieldSettingsT = None,
        exclude: Iterable[str] = None,
        encoder: EncoderT = None,
        decoder: DecoderT = None,
    ):
        self.signature_only = signature_only
        self.case = case
        self.omit = freeze(omit)  # type: ignore
        self.fields = cast(FieldSettingsT, freeze(fields)) or ()
        self.exclude = cast(Iterable[str], freeze(exclude)) or ()
        self.encoder = encoder
        self.decoder = decoder

    def merge(self, other: "SerdeFlags") -> "SerdeFlags":
        """Merge the values of another SerdeFlags instance into this one."""
        case = other.case or self.case
        signature_only = self.signature_only or other.signature_only
        if other.omit and self.omit:
            omit = (*self.omit, *(o for o in other.omit if o not in self.omit))
        else:
            omit = other.omit or self.omit  # type: ignore

        if other.fields and self.fields:
            if not isinstance(other.fields, Mapping):
                other.fields = freeze({x: x for x in other.fields})  # type: ignore
            if not isinstance(self.fields, Mapping):
                self.fields = freeze({x: x for x in self.fields})  # type: ignore
            fields = {**self.fields, **other.fields}  # type: ignore
        else:
            fields = other.fields or self.fields  # type: ignore

        if other.exclude and self.exclude:
            exclude = {*self.exclude, *other.exclude}
        else:
            exclude = other.exclude or self.exclude  # type: ignore
        encoder = other.encoder or self.encoder
        decoder = other.decoder or self.decoder
        return SerdeFlags(
            signature_only=signature_only,
            case=case,
            omit=omit,
            fields=fields,
            exclude=exclude,
            encoder=encoder,
            decoder=decoder,
        )


class SerdeConfigD(TypedDict):
    fields: Mapping[str, "AnnotationT"]
    fields_out: Mapping[str, str]
    fields_in: Mapping[str, str]
    fields_getters: Mapping[str, Callable[[str], Any]]
    omit_values: Tuple[Any, ...]
    encoder: Optional[EncoderT]
    decoder: Optional[DecoderT]


@util.slotted
@dataclasses.dataclass
class SerdeConfig:
    flags: SerdeFlags = dataclasses.field(default_factory=SerdeFlags)
    fields: Mapping[str, Annotation] = dataclasses.field(default_factory=dict)
    fields_out: Mapping[str, str] = dataclasses.field(default_factory=dict)
    fields_in: Mapping[str, str] = dataclasses.field(default_factory=dict)
    fields_getters: Mapping[str, Callable[[str], Any]] = dataclasses.field(
        default_factory=dict
    )
    omit_values: Tuple[Any, ...] = dataclasses.field(default_factory=tuple)
    encoder: Optional[EncoderT] = None
    decoder: Optional[DecoderT] = None

    def __hash__(self):
        return hash(f"{self}")

    @reprlib.recursive_repr()
    def __repr__(self) -> str:
        fs = []
        for field in dataclasses.fields(self):
            fs.append(f"{field.name}={getattr(self, field.name)!r}")
        return f"{self.__class__.__name__}({', '.join(fs)})"

    def asdict(self) -> SerdeConfigD:
        return SerdeConfigD(
            fields=self.fields,
            fields_out=self.fields_out,
            fields_in=self.fields_in,
            fields_getters=self.fields_getters,
            omit_values=self.omit_values,
            encoder=self.encoder,
            decoder=self.decoder,
        )


_AT = TypeVar("_AT")


@util.slotted(dict=False)
@dataclasses.dataclass(unsafe_hash=True)
class Annotation(Generic[_AT]):
    """The resolved, actionable annotation for a given annotation."""

    EMPTY = EMPTY

    resolved: _AT
    """The type annotation used to build the coercer."""
    origin: Type
    """The "origin"-type of the original annotation.

    Notes
    -----
    This is not necessarily the "origin"-type of the ``annotation`` attribute.
    """
    un_resolved: Any
    """The type annotation before resolving super-types."""
    parameter: inspect.Parameter
    """The parameter this annotation refers to."""
    translator: TranslatorT[_AT] = dataclasses.field(init=False)
    """A factory for generating a translation protocol between higher-level types."""
    optional: bool = False
    """Whether this annotation allows null/default values."""
    strict: st.StrictModeT = st.STRICT_MODE
    """Whether to enforce the annotation, rather than coerce."""
    static: bool = True
    """Whether we may compile a protocol ahead of time for an annotation."""
    serde: SerdeConfig = dataclasses.field(default_factory=SerdeConfig)
    """The configuration for serializing and deserializing the given type."""
    constraints: Optional["const.ConstraintsT"] = None
    """Type restriction configuration, if any."""
    generic: Type = dataclasses.field(init=False)
    has_default: bool = dataclasses.field(init=False)
    is_class_var: bool = dataclasses.field(init=False)
    resolved_origin: _AT = dataclasses.field(init=False)
    args: Tuple[Type, ...] = dataclasses.field(init=False)

    def __post_init__(self):
        self.has_default = self.parameter.default is not self.EMPTY
        self.args = util.get_args(self.resolved)
        self.resolved_origin = util.origin(self.resolved)
        self.generic = getattr(self.resolved, "__origin__", self.resolved_origin)
        self.is_class_var = isclassvartype(self.un_resolved)


_empty = object()


@util.slotted(dict=False)
@dataclasses.dataclass(unsafe_hash=True)
class ForwardDelayedAnnotation:
    ref: ForwardRef
    module: str
    resolver: "Resolver"
    parameter: Optional[inspect.Parameter] = None
    is_optional: Optional[bool] = None
    is_strict: Optional[st.StrictModeT] = None
    flags: Optional["SerdeFlags"] = None
    default: Any = _empty
    frame: Optional[FrameType] = dataclasses.field(default=None, hash=False)
    _name: Optional[str] = None
    _resolved: Optional["SerdeProtocol"] = dataclasses.field(default=None)

    @reprlib.recursive_repr()
    def __repr__(self):
        return (
            f"{self.__class__}("
            f"ref={self.ref},"
            f"module={self.module}!r, "
            f"parameter={self.parameter}, "
            f"is_optional={self.is_optional}, "
            f"is_strict={self.is_strict}, "
            f"flags={self.flags}, "
            f"default={self.default})"
        )

    @property
    def resolved(self):
        if self._resolved is None:
            globalns = sys.modules[self.module].__dict__.copy()
            try:
                type = evaluate_forwardref(self.ref, globalns, globalns)
            except NameError as e:
                name = self.ref.__forward_arg__
                type = util.extract(name, frame=self.frame)
                if not type:
                    warnings.warn(
                        f"Couldn't resolve forward reference: {e}. "
                        f"Make sure this type is available in {self.module}."
                    )
                    type = Any
            anno = self.resolver.annotation(
                type,
                name=self._name,
                parameter=self.parameter,
                is_optional=self.is_optional,
                is_strict=self.is_strict,
                flags=self.flags,
                default=EMPTY if self.default is _empty else self.default,
            )
            self._resolved = self.resolver._resolve_from_annotation(anno)
        return self._resolved

    @property
    def name(self) -> str:
        return util.get_name(self.ref)


@util.slotted(dict=False)
@dataclasses.dataclass(unsafe_hash=True)
class DelayedAnnotation:
    type: Type
    resolver: Resolver
    parameter: Optional[inspect.Parameter] = None
    is_optional: Optional[bool] = None
    is_strict: Optional[st.StrictModeT] = None
    flags: Optional["SerdeFlags"] = None
    default: Any = _empty
    _name: Optional[str] = None
    _resolved: Optional["SerdeProtocol"] = dataclasses.field(default=None)

    @reprlib.recursive_repr()
    def __repr__(self):
        return (
            f"{self.__class__}("
            f"parameter={self.parameter}, "
            f"is_optional={self.is_optional}, "
            f"is_strict={self.is_strict}, "
            f"flags={self.flags}, "
            f"default={self.default})"
        )

    @property
    def resolved(self):
        if self._resolved is None:
            anno = self.resolver.annotation(
                self.type,
                name=self._name,
                parameter=self.parameter,
                is_optional=self.is_optional,
                is_strict=self.is_strict,
                flags=self.flags,
                default=EMPTY if self.default is _empty else self.default,
            )
            self._resolved = self.resolver._resolve_from_annotation(anno)
        return self._resolved

    @property
    def origin(self):
        return self.type

    @property
    def name(self) -> str:
        return util.get_name(self.type)


AnnotationT = Union[Annotation, DelayedAnnotation, ForwardDelayedAnnotation]


class DelayedSerdeProtocol(SerdeProtocol):
    __slots__ = (
        "delayed",
        "_resolved",
    )

    def __init__(
        self,
        delayed: Union[ForwardDelayedAnnotation, DelayedAnnotation],
    ):
        self.delayed = delayed
        self.transmute = lambda val: self.deserialize(val)  # type: ignore
        self._resolved = False

    def __repr__(self):
        return f"{self.__class__.__name__}(delayed={self.delayed}, resolved={self._resolved})"

    def __delayed_init__(self):
        if self._resolved:
            return
        protocol = self.delayed.resolved
        super().__init__(
            annotation=protocol.annotation,
            constraints=protocol.constraints,
            deserialize=protocol.deserialize,
            decode=protocol.decode,
            serialize=protocol.serialize,
            encode=protocol.encode,
            validate=protocol.validate,
            translate=protocol.translate,
            iterate=protocol.iterate,
            tojson=protocol.tojson,
        )
        self._resolved = True

    def __getattr__(self, item):
        self.__delayed_init__()
        return super().__getattribute__(item)

    def __call__(self, val: Any) -> ObjectT:
        return self.transmute(val)  # type: ignore


SerdeProtocolsT = Dict[str, SerdeProtocol]
"""A mapping of attr/param name to :py:class:`SerdeProtocol`."""


class _Unprocessed:
    def __repr__(self):
        return "<unprocessed>"


Unprocessed = _Unprocessed()


class _Omit:
    def __repr__(self):
        return "<omit>"


Omit = _Omit()
KT = TypeVar("KT")
VT = TypeVar("VT")
KVPairT = Tuple[KT, VT]
